/* Copyright (C) 2015-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp.h"
#include "test_ssp_utils.h"

#include <string.h>

#define NSAMPS 1024

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct ssp_ranst_discrete* ran;
  struct ssp_rng* rng;
  const double weights[] = { 0.5, 0.1, 0.2, 0.05, 0.15 };
  const size_t nweights = sizeof(weights)/sizeof(double);
  size_t tmp[sizeof(weights)/sizeof(double)];
  double accum;
  size_t i;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);
  CHK(ssp_rng_create(&allocator, SSP_RNG_THREEFRY, &rng) == RES_OK);

  CHK(ssp_ranst_discrete_create(NULL, NULL) == RES_BAD_ARG);
  CHK(ssp_ranst_discrete_create(NULL, &ran) == RES_OK);

  CHK(ssp_ranst_discrete_ref_get(NULL) == RES_BAD_ARG);
  CHK(ssp_ranst_discrete_ref_get(ran) == RES_OK);
  CHK(ssp_ranst_discrete_ref_put(NULL) == RES_BAD_ARG);
  CHK(ssp_ranst_discrete_ref_put(ran) == RES_OK);
  CHK(ssp_ranst_discrete_ref_put(ran) == RES_OK);

  CHK(ssp_ranst_discrete_create(&allocator, &ran) == RES_OK);

  CHK(ssp_ranst_discrete_setup(NULL, NULL, 0) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(ran, NULL, 0) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(NULL, weights, 0) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(ran, weights, 0) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(NULL, NULL, nweights) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(ran, NULL, nweights) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(NULL, weights, nweights) == RES_BAD_ARG); 
  CHK(ssp_ranst_discrete_setup(ran, weights, nweights) == RES_OK); 

  memset(tmp, 0, sizeof(tmp));
  FOR_EACH(i, 0, NSAMPS) {
    const size_t k = ssp_ranst_discrete_get(rng, ran);
    double pdf;
    CHK(k < nweights);
    pdf = ssp_ranst_discrete_pdf(k, ran);
    ++tmp[k];
    CHK(pdf == weights[k]);
    CHK(pdf >= 0.f);
    CHK(pdf <= 1.f);
  }
  FOR_EACH(i, 0, nweights) {
    CHK(tmp[i] != 0);
  }

  CHK(ssp_ranst_discrete_setup(ran, weights, nweights-1) == RES_OK);
  FOR_EACH(i, 0, NSAMPS) {
    const size_t k = ssp_ranst_discrete_get(rng, ran);
    double pdf;
    CHK(k < nweights-1);
    pdf = ssp_ranst_discrete_pdf(k, ran);
    CHK(pdf >= 0.f);
    CHK(pdf <= 1.f);
  }
  accum = 0;
  FOR_EACH(i, 0, nweights-1) accum += ssp_ranst_discrete_pdf(i, ran);
  CHK(eq_eps(accum, 1, 1.e-8) == 1);
  CHK(ssp_ranst_discrete_ref_put(ran) == RES_OK);

  CHK(ssp_rng_ref_put(rng) == RES_OK);
  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);
  return 0;
}

